#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# schemeRecordToRecordType.py
# Version 1.0
# Copyright © 2019 Mark Damon Hughes
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Filters lines, any matching `(define-record TYPENAME FIELD...)`
are transformed into srfi-99 define-record-type, and read/write methods valid in Chicken.
Note I use srfi-17 setters, so don't export -set! methods.
For inheritance, change TYPENAME to `(TYPENAME PARENT)`.
Using #t creates default `make-TYPENAME` and `TYPENAME?` constructor & predicate.
For read-only, add TYPENAME-FIELD as a second list item to a field; the third item would be the setter. For example `(x)` is read/write, `(x Point-x)` is read-only,
	`(x Point-x Point-x-set!)` is also read/write.

Output template:
```
;; (define-record TYPENAME FIELD)
(define-record-type TYPENAME #t #t
	(FIELD)
)
(define-reader-ctor 'TYPENAME make-TYPENAME)
(define-record-printer (TYPENAME t out)
	(format out "#,(TYPENAME ~S)"
		(TYPENAME-FIELD t)
))
; module exports:
; make-TYPENAME TYPENAME? TYPENAME-FIELD
```
"""

import os, sys, re

def isRecordType(line):
	parseLine = line.strip()
	return parseLine.startswith("(define-record") and parseLine.endswith(")")

def recordToRecordType(line):
	parseLine = line.strip()
	if parseLine.endswith(")"):
		parseLine = parseLine[:-1]
	words = list(filter(lambda x: len(x) > 0, re.split(r"[\s]+", parseLine)))
	typename = words[1]
	typevar = typename[0].lower()
	fields = words[2:]
	s = ";; "+line+"\n"
	s += "(define-record-type %s #t #t\n" % (typename,)
	for f in fields:
		s += "\t(%s)\n" % (f,)
	s += ")\n"
	# read/write
	s += "(define-reader-ctor '%s make-%s)\n" % (typename, typename)
	s += "(define-record-printer (%s %s out)\n" % (typename, typevar)
	s += "\t(format out \"#,(%s%s)\"\n\t\t" % (typename, " ~S"*len(fields))
	for f in fields:
		s += "(%s-%s %s) " % (typename, f, typevar)
	s += "\n))\n"
	# exports
	s += "; module exports:\n"
	s += "; make-%s %s?" % (typename, typename)
	for f in fields:
		s += " %s-%s" % (typename, f)
	s += "\n"
	return s

def schemeRecordToRecordType(line):
	line = line.rstrip()
	if isRecordType(line):
		return recordToRecordType(line)
	else:
		return line

if __name__ == "__main__":
	for line in sys.stdin:
		print(schemeRecordToRecordType(line))
